/*
	This is a .QWK to local usenet news translator.
	Given that I didn't want to write a full blown .QWK reader
	when news readers are readily available, (trn, rn etc) why
	not just convert the .QWK info into a local news group
	and read them there?
	So, what this program will do is convert a series of downloaded
	articles from a FIDO style BBS that supports QWK format,
	into a usenet group with the bbsid as the top level.
	i.e. bbs name of 'PANDORA' will have a series of groups
	under 'pandora.xxx' where 'xxx' is the name of the conference
	carried at pandora, and off-line read.

	The following changes/enhancements are planned.
	1) Write some code to convert a usenet relay spool
	   into qwk format for uploading. (For replies)

	The following known bugs exist.
	1) Cannot handle more than fixed number of conferences at a time
	   (configuration parameter)

	Diane Bruce	Sept 24 '92
*/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <fcntl.h>
#include "config.h"

#define YES 1
#define NO 0

char version[]="$Id: decodeqwk.c,v 1.2 1992/11/09 03:24:02 db Exp db $";

/*
	From spec for .QWK format as outlined in news article

Article: 52 of comp.mail.misc
Xref: diana alt.bbs:27 comp.mail.misc:52 alt.bbs.internet:5
Newsgroups: alt.bbs,comp.mail.misc,alt.bbs.internet
Path: diana!gjk!revcan!cunews!utgpu!cs.utexas.edu!uunet!world!aboba
From: aboba@world.std.com (Bernard D Aboba)
Subject: Re: looking for spec for ".qwk" mailbox format
Message-ID: <BJG6y4.M83@world.std.com>
Organization: The World Public Access UNIX, Brookline, MA
References: <1992Feb8.221931.24085@msen.com>
Date: Tue, 11 Feb 1992 08:15:40 GMT
Status: OR

Here is the spec for the .QWK format. Hopefully this info will prove
useful to someone writing a QWK reader for UNIX. 

	Thank you Bernard...
*/

/*
  First record of messages.dat consists of who/what made the file.
*/

struct first_record {
  char info[128];
};

/*
  The first record of the message tells us when the message was entered
  to whom and from the message is etc. and how many more 128 byte
  blocks are in the actual message.
*/

struct message_record {
  char message_status;
  char message_number[7];
  char message_date[8];
  char message_time[5];
  char message_to[25];
  char message_from[25];
  char message_subject[25];
  char message_password[12];
  char message_reference[8];
  char message_nblocks[6];
  char message_kill;
  char conference_number[2];
  char xxx[3];	/* Not used */
};

/*
  Each successive 128 byte block consists of the message with each line
  separated by a 'SEP' character.
*/

#define SEP 227		/* Text separator character */

struct message_block{
  char message_line[128];
};

/*
  The group number to group name translation table to be built
  up by 'decode_control'

  Ideally this would be done using a hash table lookup, as this
  would enable this code to work on more than 256 groups at a time.
  However, In using this code, I have never had more than a
  few groups at a time downloaded, since the BBS I download from
  only allows 600 messages at a time to be packed.
*/

unsigned char *group_names[MAX_GROUPS];

int fin_message_in;
FILE *fp_control_in;

void decode_control();
void decode_messages();
void strtolower();
void make_usenet_date();
void strip_crlf();
void strip_trailing_blanks();
void safe_fgets();
void bad_control();
void bad_message();


unsigned char *cut_to_length();
unsigned char *map_to_news_name();
int get_group_number();

int getopt();

char *make_legal_group_name();
char *find_fido_address();

int test;

int number_of_conferences;
char organization[MAX_IN_LINE];
char bbs_id[MAX_IN_LINE];

main(argc,argv)
int argc;
char *argv[];
{
  int c;
  extern char *optarg;
  extern int optind, opterr;
  test = NO;

/*
  If -t option is set, then decoded messages go directly to standard out
  If -v option is set, version is printed, and program exits.
  I always check for '-?' or '-h' for help.
*/

  opterr = 0;

  while ((c = getopt(argc, argv, "vth?")) != -1)
    {
      switch (c)
	{
	case 't':
	  test = YES;
	  break;
	case 'v':
	  (void)fprintf(stderr,"%s: version %s\n", argv[0], version );
	  (void)exit(0);
	  break;
	case 'h':
	case '?':
	default:
	  (void)fprintf(stderr,"%s [-t][-v]\n", argv[0]);
	  (void)fprintf(stderr,
	  "-t for 'test' mode, decoded messages output to standard output.\n");
	  (void)fprintf(stderr,"-v for 'version'.\n");
	  (void)fprintf(stderr,"-? -h for help.\n");
	  (void)exit(0);
	  break;
	}
    }

  if( (fin_message_in = open("messages.dat", O_RDONLY)) < 0 )
    {
      (void)fprintf(stderr,"%s: Unable to open \"messages.dat\"\n", argv[0]);
      (void)exit(0);
    }

  if( (fp_control_in = fopen("control.dat", "r")) == NULL )
    {
      (void)fprintf(stderr,"%s: Unable to open \"control.dat\"\n", argv[0]);
      (void)exit(0);
    }

  decode_control( fp_control_in );
  decode_messages( argv[0], fin_message_in );
  (void)exit(0);
}

/*
  decode_control()

  Decode the contents of the control.dat file.

  What I have to do here is:
    1)	I determine what the originating BBS is for this qwk file.

    2)	Build up a decode table of numbers to Conference names,
	as found in the control.dat file. These will be used to
	produce the correct local newsgroup name when the
	messages.dat file is decoded.
	To simplify things, I have assumed a minimum number of
	groups are available i.e. less than MAX_GROUPS. The qwk
	format seems to allow up to 65535 groups, but I should think
	256 is probably plenty.
	
*/

void decode_control(fp)
FILE *fp;
{
char line_in[MAX_IN_LINE];
int group_number;
int i,j;

  safe_fgets(organization, MAX_IN_LINE, fp, bad_control );

  for( i = 0; i < MAX_IN_LINE; i++ )
    if( (organization[i] == '\n') || (organization[i] == '\r' ) )
      organization[i] = '\0';

	/*
	  Skip the unused records
	*/

  safe_fgets(line_in, MAX_IN_LINE, fp, bad_control);	/* blank line */
  safe_fgets(line_in, MAX_IN_LINE, fp, bad_control);	/* phone number */
  safe_fgets(line_in, MAX_IN_LINE, fp, bad_control);	/* SYSOP */
  safe_fgets(line_in, MAX_IN_LINE, fp, bad_control);
					/* Mail door reg#,BBS name */
	
  for( j = i = 0; i < MAX_IN_LINE; i++ )
    {
      if( line_in[i] == ',' )
	{
	  j = 0;
	  continue;
	}
      else if( (line_in[i] == '\n') || (line_in[i] == '\r') )
	{
	  line_in[i] = '\0';
	  break;
	}
      else if( line_in[i] == ' ' )
	line_in[i] = '\0';
      bbs_id[j++] = line_in[i];
    }
  strtolower(bbs_id);

/*
  Discard whatever lines are not needed.
*/
  for( i = 0; i < 5; i++ )
    {
      safe_fgets(line_in, MAX_IN_LINE, fp, bad_control);
    }


  safe_fgets(line_in, MAX_IN_LINE, fp, bad_control);
  (void)sscanf(line_in,"%d", &number_of_conferences);
  number_of_conferences++;

  for( i = 0; i < number_of_conferences; i++ )
    {
      safe_fgets(line_in, MAX_IN_LINE, fp, bad_control);
      (void)sscanf(line_in,"%d",&group_number);

      safe_fgets(line_in, MAX_IN_LINE, fp, bad_control);

      if( (group_number < 0) || (group_number > MAX_GROUPS) )
	group_number = 0;

      strip_crlf( line_in );
      strip_trailing_blanks( line_in );
      group_names[ group_number ] = (unsigned char*)strdup( line_in );
    }
}

/*
  strip_crlf()

  What I am doing here:
    1) Discarding the MSDOS <CR><LF> at the end of each line, in place.
*/

void strip_crlf( line_in )
char *line_in;
{
int j;

/*
  If it's MSDOG it should always be <CR><LF>, so I could always be lazy
  and assume the <CR> is first... ha!
*/
  for( j = 0; j < MAX_IN_LINE; j++ )
    {
      if( ( line_in[j] == '\n') || (line_in[j] == '\r') )
	{
	  line_in[j] = '\0';
	  break;
	}
    }
}

/*
  strip_trailing_blanks()
  
  What I am doing here:
  1) Simply replacing trailing blank characters with a null,
     in place, thus truncating the string.
*/

void strip_trailing_blanks( line_in )
char *line_in;
{
int j;

  for( j = 0; j < MAX_IN_LINE; j++ )
    {
      if( isspace( line_in[j] ) )
	{
	  line_in[j] = '\0';
	  break;
	}
    }
}

/*
  make_legal_group_name()

  What I am doing here:
    1) replacing embedded blanks with an underscore
    2) replacing embedded '/' with a period character
    3) adding bbs name to beginning of name
    4) convert the whole mess to lower case
    5) return a pointer to this new legal name.
       Which is in a static, and must be copied immediately.
*/

char *make_legal_group_name( line_in )
char *line_in;
{
int j;
static char legal_name[2*MAX_LEGAL_NAME];

  for( j = 0; j < MAX_IN_LINE; j++ )
    {
      if( line_in[j] == ' ' ) line_in[j] = '_';
      else if( line_in[j] == '/' ) line_in[j] = '_';
      else if( line_in[j] == '\0' )break;
    }

  legal_name[0] = '\0';

  (void)strcpy(legal_name,bbs_id);
  (void)strcat(legal_name,".");

  (void)strcat(legal_name, line_in );
  strtolower( legal_name );
  return(legal_name);
}

/*
  strtolower()
  
  1) Convert a string in place from upper case to lower case.
*/

void strtolower( s )
char *s;
{
  while( *s )
    {
      if( isupper(*s) ) *s = tolower(*s);
      s++;
    }
}

/*
  decode_messages()

  1) "Walk" through the 128 byte records (ugh! shades of CPM!)
     as defined by the .QWK standard. 
*/

void decode_messages(prog_name,fin)
char *prog_name;
int fin;
{
struct first_record info_record;
struct message_record msg_info_record;
struct message_block msg_record;
int nblocks;
int i;
unsigned char *p;
char temp_date[20];
FILE *out;
FILE *temp_out;
char line_in[MAX_IN_LINE];
char template[40];
char *fido_address;
int group_number;
int being_mailed;

  if( test )
    out = stdout;

  if( read(fin,(char *)&info_record, sizeof( info_record )) <= 0 )
    {
      (void)fprintf(stderr,"Malformed messages.dat, bailing out!\n");
      (void)exit(0);
    }

/*
  If one is curious as to who the messages were packed by...

  p = cut_to_length(info_record.info,128);
  (void)fprintf(stderr,"Messages packed by %s\n", p );
*/

  while(read(fin,(char *)&msg_info_record, sizeof( msg_info_record)) > 0)
    {
      being_mailed = NO;
      (void)strcpy(template,"/tmp/decodeqwkXXXXXX");
      temp_out = fdopen(mkstemp(template),"r+");
      if( temp_out == NULL )
	{
	  (void)fprintf(stderr,"Can't open temp_out\n");
	  (void)exit(0);
	}

/*
  If I am debugging the output, I do not actually call up
  sendmail or relaynews, but output to stdout.
*/

      if( !test )
	{
/*
  Private message has '*' or '-' set in the status field,
  I am being simple minded and assuming that I will not see
  a 'To' field that isn't on my private unix machine,
  if is a private message. i.e. a private message obviously
  wouldn't be sent to me if it belongs to someone else.
  N.B. An alias used on the BBS end will be translated to the 
  same alias name on the unix end however...
*/
	  if( msg_info_record.message_status == '*' ||
	     msg_info_record.message_status == '+' )
	    {
/*
  I redundantly get the 'to' field if it is a private message. so?
*/
	      p = cut_to_length(msg_info_record.message_to, 25 );
	      (void)strcpy(line_in,SENDMAIL);
	      (void)strcat(line_in," ");
	      (void)strcat(line_in, (char *)p);
	      being_mailed = YES;
	      out = popen(line_in,"w");
	      if( out == (FILE *)NULL )
		{
		  (void)fprintf(stderr,"%s: Cannot mail\n", prog_name);
		  (void)exit(0);
		}
	    }
	  else
	    {
	      out = popen(RELAYNEWS,"w");
	      if( out == (FILE *)NULL )
		{
		  (void)fprintf
		    (stderr,"%s: Cannot run relaynews\n", prog_name);
		  (void)exit(0);
		}
	    }
	}

      p = cut_to_length(msg_info_record.message_from, 25 );
      (void)fprintf(out,"From: %s\n", p );
      (void)fflush(out);

      p = cut_to_length(msg_info_record.message_to, 25 );
      (void)fprintf(out,"To: %s\n", p );
      (void)fflush(out);

      group_number = get_group_number(msg_info_record.conference_number);
      p = map_to_news_name( group_number );
      (void)fprintf(out,"Newsgroups: %s\n", p );
      (void)fprintf(out,"Path: diana!%s\n", bbs_id);
      p = cut_to_length(msg_info_record.message_subject, 25 );
      (void)fprintf(out,"Subject: %s\n", p );

    /*
      Date needs some massaging to be a legal Usenet date
      legal date needs day# month year# with month spelled
      out in 3 letter short form
    */
      make_usenet_date(msg_info_record.message_date,temp_date);
      (void)fprintf(out,"Date: %s ", temp_date);
    /*
      Output the time, show the seconds as zero since it isn't
      supplied but the field is needed
    */

      p = cut_to_length(msg_info_record.message_time, 5 );
      (void)strcat((char *)p,":00\n");
      (void)fputs((char *)p,out);

      (void)sscanf(msg_info_record.message_nblocks,"%6d",&nblocks);
      nblocks--;

      while ( nblocks )
	{
	  if( read(fin,(char *)&msg_record, sizeof( msg_record )) <= 0 )
	     {
	       (void)fprintf(stderr,"Malformed messages.dat, bailing out!\n");
	       (void)exit(0);
	     }

	  p = (unsigned char *)&msg_record;
	  for( i = 0 ; i < 128; i++ )
	    if( p[i] == SEP )
	      (void)fputc('\n',temp_out);
	    else
	      (void)fputc((char)p[i],temp_out);
	  nblocks--;
	}
      (void)fputc('\n',temp_out);
      
      (void)rewind( temp_out );

      fido_address = NULL;
      
      while( fgets(line_in,MAX_IN_LINE,temp_out) != NULL )
	{
	  if(strncmp(line_in+1,"* Origin:",9) == 0 )
	    {
	      (void)strcpy(organization, line_in+11 );
	      strip_crlf( organization );

	      fido_address = find_fido_address(organization);
	      
	      if( fido_address == NULL )
		{
		  if( fgets(line_in,MAX_IN_LINE,temp_out) != NULL )
		    {
		      strip_crlf( line_in );
		      fido_address = find_fido_address(line_in);
		    }
		  break;
		}
	    }
	}

      if( fido_address == NULL )
	fido_address = bbs_id;

      if( !being_mailed )
	{
	  short unsigned int checksum;
/*
  The message id is derived from as much unique information as I could
  gather from the message headers. In the FIDO spec, there is
  a unique message ID used to prevent infinite message loops
  much as 'Message-Id"' is used in the usenet world. (It is hidden
  using a Control-A in the text)
  However, it does not appear in the QWK format output, so I fake one up.

  The time/date/group name and fido address are almost enough to be
  unique, unfortunately sometimes two messages still 'hash' to
  same Id, (Multi-ported BBS's or fast message entry.) so
  I 'perturb' the Id using a very simple checksum.

  I did use the message number, but this meant that the identical
  same message downloaded from another board would have
  a different ID.
  The only solution I could think of is a checksum calculated
  from the contents or the header of the message.

  A very simple 16 bit checksum of the to/from and subject
  fields of the message should be enough.
*/
	  checksum = 0;

	  for(i = 0; i < 25; i++ )
	    checksum += msg_info_record.message_to[i];
	  for(i = 0; i < 25; i++ )
	    checksum += msg_info_record.message_from[i];
	  for(i = 0; i < 25; i++ )
	    checksum += msg_info_record.message_subject[i];

	  (void)fprintf(out,
			"Message-Id: <19%c%c%c%c%c%c%c.%c%c%c%c%d.%s@%s>\n",
			temp_date[7],temp_date[8],
			temp_date[3],temp_date[4],temp_date[5],
			temp_date[0],temp_date[1],
			msg_info_record.message_time[0],
			msg_info_record.message_time[1],
			msg_info_record.message_time[3],
			msg_info_record.message_time[4],
			checksum,
			group_names[group_number],
			fido_address);
	      }

      (void)fprintf(out,"Organization: %s\n\n", organization );
      
      rewind( temp_out );
      while( fgets(line_in,MAX_IN_LINE,temp_out) != NULL )
	{
	  (void)fputs(line_in, out );
	}
      (void)fclose( temp_out );
      (void)unlink(template);

      if( !test )
	(void)pclose( out );
    }
  (void)fclose( temp_out );
  (void)unlink(template);
}

/*
  make_usenet_date()

  Convert fidonet date form into usenet date form
*/

char *month_strings[]={"???","Jan","Feb","Mar","Apr","May","Jun",
			     "Jul","Aug","Sep","Oct","Nov","Dec"};

void make_usenet_date( fidonet_date, usenet_date )
char *fidonet_date;
char *usenet_date;
  {
    int date_number;

/*       I know, it's quick and dirty */

    (void)strcpy(usenet_date,"         ");	/* blank fill it */

    /* Put the day number into right spot */
    usenet_date[0] = fidonet_date[3];
    usenet_date[1] = fidonet_date[4];
    /* Put the year number into right spot */
    usenet_date[7] = fidonet_date[6];
    usenet_date[8] = fidonet_date[7];
    /* Getting month converted is trickier */
    (void)sscanf(fidonet_date,"%d",&date_number);
    if( date_number > 12 || date_number < 0 ) date_number = 0;
    (void)strncpy(usenet_date+3,month_strings[date_number],3);
    usenet_date[9] = '\0';
  }

/*
  cut_to_length()

    1) Cut a string down to maximum length as given by length parameter
       blow away trailing blank space as it is a waste of space.
    2) return a pointer to this new string.

*/

unsigned char *cut_to_length(s,length)
char *s;
int length;
{
static unsigned char temp_buff[MAX_IN_LINE];
int i;

  if( length >= MAX_IN_LINE )length = MAX_IN_LINE - 1;
  (void)strncpy((char *)temp_buff,s,length);
  temp_buff[length] = '\0';
  i = length - 1;

  while( i > 0 )
    {
      if( isspace(temp_buff[i]) )
	temp_buff[i--] = '\0';
      else
	break;
    }
  return( temp_buff );
}

/*
  find_fido_address()
  
    1) Try to figure out what the fido net address is...

    Form of fido net address should be "(xx:xxxx/x)"
    There must be a ':' found, or I could be seeing an area
    code (*sigh*), i.e. some addresses have (xxx) xxx-xxxx
    for phone number...
*/

char *find_fido_address( s )
char *s;
{
int i,j;
static char fido_address[40];
int colon_found = NO;

  if( s == NULL )
    i = 0;
  else
    i = strlen( s );

  if( i != 0 ) i--;

  while( i > 0 )
    {
      if( s[i-1] == '(' )break;
      if( s[i] == ':' )colon_found = YES;
      i--;
    }

  if( (i == 0) || !colon_found )  /* Didn't find a valid fido address, */
      return( NULL );             /* try next line */

  j = 0;
  while( (s[i] != ')') && (s[i] != '\0'))
    {
      if( isspace( s[i] ) ) s[i] = '_';
      fido_address[j++] = s[i++];
      }
  fido_address[j] = '\0';

  return(fido_address);
}

/*
  map_to_news_name()

  1) map a group number into a group name.
*/

unsigned char *map_to_news_name( number_in )
int number_in;
{
  return((unsigned char *)
	 make_legal_group_name ( (char *)group_names[number_in]));
}

/*
  get_group_number

  1) Convert two 1 byte binary values to one integer, return this value as
  int. Done this way to get around little endian versus big endian problems.
*/

int get_group_number( char_number_in )
char char_number_in[2];
{
int number;

  number = ((int)char_number_in[0] & 0xFF);
/*	number += (((int)char_number_in[1] & 0xFF) * 256); */
  return( number );
}

/*
  safe_fgets
    exit with appropriate error message if a fgets returns
    an empty line at the wrong time.
*/

void safe_fgets( line_in, max, fp, error_fn )
char *line_in;
int max;
FILE *fp;
void (*error_fn)();
{
      if ( fgets(line_in, max, fp) == NULL )(*error_fn)();
}

void bad_control()
  {
    (void)fprintf(stderr,"Malformed control.dat, bailing out!\n");
    (void)exit(0);
  }
